/*
- Tablettenwecker -
Klingelt 10h nach einlegen der Batterie.
Dann tglich um die gleiche Zeit.
Klingeln wird auf Tastendruck oder automatisch 
nach 20 Piepsern beendet.

Extremes stromsparen:
  - Uhrenquarz an Timer T2; weckt den Prozessor alle 8s
  - Takt auf 128KHz
  - IRQ-Routine mit handoptimiertem Assemblercode
  
  alles Abschalten, was nicht gebraucht wird:
  - ADC
  - Komparator
  - BrownOut-Detektor
  - Debug-Schnittstelle (JTAG)
  
  Pull-up Widerstnde auf allen Eingabeports einschalten, damit die nicht "flattern"
  Pull-up Widerstnde am Hex-Drehschalter Port nur 1x am Tag alle einschalten, sonst nur die,
  die nicht auf Masse gezogen werden.
  
Theoretische Laufzeit mit CR2032-Lithium Batterie bei 1.8V: ca. 30 Jahre
Da wir anfangs 3.0V haben, drfte die Laufzeit deutlich darunter liegen.


Realisiert mit AtMega324P
    interner 128KHz Oszillator

Pinbelegung:
    TOSC1 + TOSC2: 32768Hz Quartz
    PB3 und PB4: Piezo;
    PA0: Taster gegen GND
    PD0..6: Zeitkorrektur Wert
    PD7: Zeitkorrektur Vorzeichen L=vorstellen H=rckstellen


*/

#define F_CPU 128000UL	/* CPU clock in Hertz */

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>


#define U08 uint8_t
#define U16 uint16_t

//--------------------------------------------------------------------------
//Alle 8s wacht der Prozessor auf. 
//Dann dreht er die Uhr weiter und legt sich wieder hin.
//Damit das mglichst schnell geht, ist der Code auf Ausfhrungszeit
//von Hand optimiert, indem mglichst viel in Registern gehalten wird.
//(Ich hatte einfach keinen Bock, den Kram komplett in Assembler zu schreiben)

//Die paar Variablen werden fest Registern zugewiesen
//R0: Status in IRQ fr T1
//R2: Status in IRQ fr T2

register unsigned char tic asm("r3");
register unsigned char s8 asm("r4");        //Wird alle 8s runtergezhlt
register unsigned char m20 asm("r5");       //Wird alle 20m runtergezhlt
register unsigned char neuerTag asm("r6");  //Wird am Anfang jedes Tages gesetzt
register unsigned char sTag asm("r7");      //s8 Startwert am Tagesanfang    
register unsigned char mAlarm asm("r8");    //m20 Zhlerstand, zu dem der Alarm erfolgt
register unsigned char subKorrektur asm("r9");
register unsigned char piepszaehler asm("r10");

//Die paar Konstanten auch
register unsigned char _20 asm("r12");
register unsigned char _72 asm("r13");
register unsigned char _150 asm("r14");


register unsigned char hexschalter asm("r15");





//--------------------------------------------------------------------------
inline void schlafen(void){
  SMCR = _BV(SM1)|_BV(SM0)|_BV(SE); //Power-Save-Mode und SE enable
  asm("SLEEP");
  //Schlft bis T2 berluft (also bis zu 8s)
}

//--------------------------------------------------------------------------
//ISR(TIMER1_COMPA_vect){  //25Hz timer
void  __attribute__ ((naked)) TIMER1_COMPA_vect(void){
  asm("in	r0, 0x3f"  );    //Status sichern
  asm("inc r3"       );    //tic++;
  asm("out	0x3f, r0");    //Status wiederherstellen
  asm("reti         ");
}
//--------------------------------------------------------------------------
//ISR(TIMER2_OVF_vect){ //8s Interrupt 
void  __attribute__ ((naked)) TIMER2_OVF_vect(void){
  asm("in	r2, 0x3f"  );    //Status sichern

  //Uhr weiterdrehen (Count-Down): 10800 Interrupts am Tag, 450 pro Stunde
  s8--;
  if (!s8){ //20 Minuten sind rum
    s8=_150;
    m20--;
    if (!m20){ //Tag ist rum
      m20=_72;
      neuerTag=_72;  //Irgendwas ungleich Null
      s8=sTag;        
    }
  }
  asm("out	0x3f, r2");    //Status wiederherstellen
  asm("reti         ");
}


//--------------------------------------------------------------------------
//--------------------------------------------------------------------------
int main(void){

  cli();
  ADCSRA=0; //A/D Wandler aus
  ACSR=_BV(ACD);  //Analog Comparator disable

  //Watchdog tten
  asm("wdr");// watchdog reset
  MCUSR &= ~(1<<WDRF); // Clear WDRF in MCUSR 
  /* Write logical one to WDCE and WDE */
  /* Keep old prescaler setting to prevent unintentional time-out */
  WDTCSR |= (1<<WDCE) | (1<<WDE);
  WDTCSR = 0; /* Turn off WDT */

  MCUCR = 0x80; //JTD=1: Jtag Debug disable (muss 2x geschrieben werden)
  MCUCR = 0x80; 

  //Port Direction: 1=Output 0=Input
  DDRA=0;
  DDRB=0x18;
//  DDRC=3;  
  DDRC=0;  
  DDRD=0;

  PORTA=0xff; //Pull-Ups einschalten
  PORTB=0xff; //Pull-Ups einschalten
  PORTC=0xff; //Pull-Ups einschalten
  PORTD=0xff; //Pull-Ups einschalten

  //Set up the 8-bit timer 0  (1200Hz Beeper)
  TCCR0A = _BV(WGM01)  //CTC-Mode
        | _BV( COM0A0) //Toggle OC0A on Compare Match
        | _BV( COM0B0);//Toggle OC0B on Compare Match  
  TCCR0B = _BV(CS00) | _BV(FOC0A);  // Clock Select 001 => Teiler=1 (gibt 128KHz)
  OCR0A = 20;
  OCR0B = 20;


  //Set up the 16-bit timer 1  25Hz Zeitgeber
  TCCR1A = 0;
  TCCR1B = _BV(WGM12) 
         | _BV(CS12) | _BV(CS10); //Teilt durch 1024 - gibt 128Hz
  OCR1A=4; //gibt rund 25Hz;
  TIMSK1 = _BV(OCIE1A);
 
  //Set up the 8-bit timer 2  (32768Hz RealTime Clock)
  //--------------------------------------------------
  TCCR2A =  0;
  TCCR2B =  
            _BV(CS22)  
          | _BV(CS21)  // Clock Select 111 => Teiler=1024 (gibt 32Hz)
          | _BV(CS20) 
          ;
  ASSR = _BV(AS2); //Asyncron Real-Time-Clock; EXCLK=0

  //Timer Interrupt Mask
  TIMSK2 = _BV(TOIE2);  // Timer2 Overflow Interrupt aktivieren


  _20=20;
  _72=72;
  _150=150;
  tic=0;




  //Piepse , bis Taste gedrckt (max. 20 mal)
  sei();
  piepszaehler=0;
  while((PINA & 1) && (piepszaehler < _20)) {

    //Kurze Beeps - lngere Pausen, ansteigende Frequenz von Beep zu Beep
    if (tic > _20){
      tic=0;

      piepszaehler++;

      //Ton erhhen
      OCR0A-=1;
      if (OCR0A < 15){
        OCR0A=15;
      }
      DDRB=0x18;
      if (piepszaehler>10){
        OCR0B = OCR0A;
        TCCR0B = _BV(CS00);  // Clock Select 001 => Teiler=1 (gibt 128KHz)
      
        if (
             ((PINB & 0x08) && (PINB & 0x10))
           ||(!(PINB & 0x08) && !(PINB & 0x10))
           ){//Ausgnge sind gleich - schalte einen um (lautes Piepsen)
          TCCR0B = _BV(CS00) | _BV(FOC0A);  // 128KHz; einen Ausgang toggln
        }else{
          TCCR0B = _BV(CS00);  // Clock Select 001 => Teiler=1 (gibt 128KHz)
        }
      }else{ //leise Piepsen
        OCR0B = 255;
        TCCR0B = _BV(CS00);  // Clock Select 001 => Teiler=1 (gibt 128KHz)
      }

    }

    if (2==tic){
      TCCR0B = _BV(FOC0A);  // Timer aus, Ausgnge gleichschalten
      TCNT0=0;
      DDRB=0;     //Port auf Eingang und
      PORTB=0xff; //Pull-ups ein 
    }
  }//while

  
  PORTD=0xff; //Alle Pull-ups ein
  asm("nop");
  asm("nop");
  hexschalter = PIND ^ 0xff;
  subKorrektur = (hexschalter & 0x0f);

  if (hexschalter & 0x80){ //Uhr ggf. vorstellen
    sTag = _150 - ((hexschalter & 0x7f) >> 4);    //0..7 * 8s = 0..56s vorstellen
  }else{
    //Uhr ggf. zurueckstellen
    sTag = _150 + ((hexschalter & 0x7f) >> 4);    //0..7 * 8s = 0..56s vorstellen
  }
  PORTD = PIND; //Pull Up nur auf Eingngen, die nicht nach Masse gezogen werden

  
  //Uhr auf Anfang
  cli();
    s8=sTag;
    m20=_72;
    mAlarm=42; //72-10*3; 10 Stunden nach "Tagesanfang" (also nach Batterie einlegen)

//Test:
//    mAlarm=69; //72-3; 1 Stunde nach "Tagesanfang" (also nach Batterie einlegen)
    piepszaehler=0;
  sei();
 
  
  
  //Piepse so oft, wie halbe Sekunden Tageskorrektur
  DDRB=0x18;
  if (PIND & 0x80){ //Uhr soll langsamer (default)
    OCR0A=80;
  }else{
    OCR0A=5;  //Uhr soll schneller
  }
  OCR0B=111; //Ausgang B toggelt nicht mehr!
  tic=0;
  piepszaehler=0;
  //korrektur = PIND ^ 0xff;
  while(piepszaehler <= ((PIND ^ 0xff) & 0x1f)){ //Vorzeichen ignorieren
    if (tic>12){
      tic=0;
      //Sound einschalten
      TCCR0B = _BV(CS00);  // 128KHz
      piepszaehler++;
    }
    if (2==tic){
      TCCR0B = 0;  // Timer aus, 
      TCNT0 = 0;
    }

  }

  TCCR0B = 0;  // Clock Select 000 => Timer aus (=Piepser aus)
  DDRB=0;     //Port auf Eingang und
  PORTB=0xff; //Pull-ups ein 

  //-------------------------------------------------------
  //   Hauptschleife
  //-------------------------------------------------------
  while(1){ //Arbeite bis die Batterie rausgenomen wird

    schlafen(); //8s nix tun. T2 weckt via Interrupt.

#if 0
//Test:
//alle 20 Minuten klicken
    if (s8==_150){ 
      tic=0;
      DDRB=0x18;   //Port fr Piezo auf Ausgang
      OCR0A = 13;  //Tonhhe
      OCR0B = 255; //Toggelt nicht - piepst damit leiser
      TCCR0B = _BV(CS00);  // Clock Select 001 => Teiler=1 (gibt 128KHz)

      while (tic < 8){

      }
      TCCR0B = _BV(FOC0A);  // Timer aus, Ausgnge gleichschalten
      TCNT0=0;
      DDRB=0;     //Port auf Eingang und
      PORTB=0xff; //Pull-ups ein 
  
    }
#endif

    if (mAlarm == m20){ 
      // --- Dieses ist der  Alarm-Zeitpunkt (20 mintige "Alarm-Stunde") ---
      tic=21;
      piepszaehler=0;
      DDRB=0x18;   //Port fr Piezo auf Ausgang
      OCR0A = 20;  //Anfangstonhhe
      OCR0B = 255; //Toggelt nicht - piepst damit leiser
       
      //Bei Tastendruck oder nach 20 Piepsern aufhren zu piepsen
      while((PINA & 1) && (piepszaehler < _20)) {

        //Kurze Beeps - lngere Pausen, ansteigende Frequenz von Beep zu Beep
        if (tic > _20){
          tic=0;

          piepszaehler++;

          //Ton erhhen
          OCR0A-=1;
          if (OCR0A < 15){
            OCR0A=15;
          }
          DDRB=0x18;
          if (piepszaehler > 10){ //laut Piepsen (+3V / -3V am Piezo)
            OCR0B = OCR0A;
            TCCR0B = _BV(CS00);  // Clock Select 001 => Teiler=1 (gibt 128KHz)
  
            if (
                 ((PINB & 0x08) && (PINB & 0x10))
               ||(!(PINB & 0x08) && !(PINB & 0x10))
               ){//Ausgnge sind gleich - schalte einen um
              TCCR0B = _BV(CS00) | _BV(FOC0A);  // 128KHz; einen Ausgang toggln
            }else{
              TCCR0B = _BV(CS00);  // Clock Select 001 => Teiler=1 (gibt 128KHz)
            }
          }else{ //leise Piepsen (+3V / 0V am Piezo)
            OCR0B = 255;
            TCCR0B = _BV(CS00);  // Clock Select 001 => Teiler=1 (gibt 128KHz)
          }

        }

        if (2==tic){
          TCCR0B = _BV(FOC0A);  // Timer aus, Ausgnge gleichschalten
          TCNT0=0;
          DDRB=0;     //Port auf Eingang und
          PORTB=0xff; //Pull-ups ein 
        }

      }//Tastendruck

      //Piepser aus. Ports auf stromsparen.
      TCCR0B = _BV(FOC0A);  // Timer aus, Ausgnge gleichschalten
      TCNT0=0;
      DDRB=0;     //Port auf Eingang und
      PORTB=0xff; //Pull-ups ein 


      //Hex-Drehschalter auslesen und damit Interrupts pro Tag einstellen

      //Zeitkorrektur
      PORTD=0xff; //Alle Pull-ups ein
      asm("nop");
      asm("nop");
      hexschalter = PIND ^ 0xff;
      PORTD = PIND; //Pull Ups nur auf Eingngen, die nicht nach Masse gezogen werden
      subKorrektur += (hexschalter & 0x0f);
    
      if (hexschalter & 0x80){ //Uhr ggf. vorstellen
        sTag = _150 - ((hexschalter & 0x7f) >> 4);    //0..7 * 8s = 0..56s vorstellen
        if (subKorrektur > 15){
          sTag--;  
          subKorrektur-=16;  
        }            
      }else{
        //Uhr ggf. zurueckstellen
        sTag = _150 + ((hexschalter & 0x7f) >> 4);    //0..7 * 8s = 0..56s vorstellen
        if (subKorrektur > 15){
          sTag++;  
          subKorrektur-=16;  
        }            
      }


      while (mAlarm == m20){ //Warte, bis die Alarm-Stunde rum ist (dann ist der Alarm wieder scharf)
        schlafen();
      }

    }//if (mAlarm 
    if (neuerTag){
      neuerTag=0;
      tic=0;
      DDRB=0x18;   //Port fr Piezo auf Ausgang
      OCR0A = 13;  //Tonhhe
      OCR0B = 255; //Toggelt nicht - piepst damit leiser
      TCCR0B = _BV(CS00);  // Clock Select 001 => Teiler=1 (gibt 128KHz)

      while (tic < 15){

      }
      TCCR0B = _BV(FOC0A);  // Timer aus, Ausgnge gleichschalten
      TCNT0=0;
      DDRB=0;     //Port auf Eingang und
      PORTB=0xff; //Pull-ups ein 
    }

  }//While 1
}

